# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests
import testtools

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import java
from soufi.testing import base


class TestJavaFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return java.JavaFinder(name, version, SourceType.java)

    def test_get_source_url(self):
        finder = self.make_finder()
        group = self.factory.make_string()
        url = self.factory.make_url()
        data = dict(response=dict(docs=[dict(g=group)]))

        get = self.patch_get_with_response(requests.codes.ok, json=data)
        head = self.patch_head_with_response(requests.codes.ok)
        head.return_value.url = url
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        expected = dict(
            q=f'a:{finder.name} v:{finder.version} l:sources', rows=1
        )
        get.assert_called_once_with(
            java.MAVEN_SEARCH_URL, params=expected, timeout=finder.timeout
        )
        expected = dict(
            filepath=f'{group}/{finder.name}/{finder.version}/'
            f'{finder.name}-{finder.version}-sources.jar'
        )
        head.assert_called_once_with(
            java.MAVEN_REPO_URL,
            params=expected,
            allow_redirects=True,
            timeout=finder.timeout,
        )

    def test_get_source_info_raises_when_get_response_fails(self):
        self.patch_get_with_response(requests.codes.bad)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_get_source_info_raises_when_head_response_fails(self):
        group = self.factory.make_string()
        data = dict(response=dict(docs=[dict(g=group)]))
        self.patch_get_with_response(requests.codes.ok, json=data)
        self.patch_head_with_response(requests.codes.not_found)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_get_source_info_raises_when_head_response_times_out(self):
        group = self.factory.make_string()
        data = dict(response=dict(docs=[dict(g=group)]))
        self.patch_get_with_response(requests.codes.ok, json=data)
        head = self.patch(requests, 'head')
        head.side_effect = requests.exceptions.Timeout
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_find(self):
        url = self.factory.make_url()
        finder = self.make_finder()
        self.patch(finder, 'get_source_url').return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, java.JavaDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test_get_release_history(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'response': {
                'docs': [
                    {'v': '2.0.0', 'timestamp': 2000},
                    {'v': '1.0.0', 'timestamp': 1000},
                ]
            }
        }
        self.patch_get_with_response(requests.codes.ok, json=data)

        history = finder.get_release_history()
        self.assertEqual(
            ['1.0.0', '2.0.0'], [item['version'] for item in history]
        )

    def test_get_release_history_sorts_none_timestamp_as_oldest(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'response': {
                'docs': [
                    {'v': '4.0.2'},
                    {'v': '5.1.0.RELEASE', 'timestamp': 1514764800000},
                ]
            }
        }
        self.patch_get_with_response(requests.codes.ok, json=data)

        history = finder.get_release_history()
        self.assertEqual(
            ['4.0.2', '5.1.0.RELEASE'],
            [item['version'] for item in history],
        )

    def test_get_release_history_raises_when_request_fails(self):
        finder = self.make_finder(name=self.factory.make_string())
        self.patch_get_with_response(requests.codes.bad_request)
        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )

    def test_get_release_history_raises_when_docs_filtered(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'response': {
                'docs': [
                    {'v': ''},
                    {'v': ''},
                ]
            }
        }
        self.patch_get_with_response(requests.codes.ok, json=data)
        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )

    def test_get_release_history_uses_group_id(self):
        name = self.factory.make_string()
        group_id = self.factory.make_string()
        finder = java.JavaFinder(
            name,
            self.factory.make_string(),
            SourceType.java,
            group_id=group_id,
        )
        data = {'response': {'docs': [{'v': '1.0.0', 'timestamp': 1000}]}}
        get = self.patch_get_with_response(requests.codes.ok, json=data)

        finder.get_release_history()
        call_args = get.call_args
        query = call_args.kwargs['params']['q']
        self.assertIn(f'g:{group_id}', query)
        self.assertIn(f'a:{name}', query)


class TestNPMDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return java.JavaDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        jds = self.make_discovered_source(url)
        self.assertEqual(url, repr(jds))

    def test_make_archive(self):
        jds = self.make_discovered_source()
        self.assertEqual(jds.make_archive, jds.remote_url_is_archive)



# Copyright (c) 2024 Cisco Systems, Inc. and its affiliates
# All rights reserved.


from soufi import exceptions, finder

DEFAULT_INDEX = "https://index.crates.io/"
CRATES_API = "https://crates.io/api/v1/crates/"


class CrateFinder(finder.SourceFinder):
    """Find Rust Crates.

    Traverses the supplied index, defaulting to the one at index.crates.io.

    :param index: optional index server; defaults to
        https://index.crates.io/
    """

    distro = finder.SourceType.crate.value

    def __init__(self, *args, **kwargs):
        self.index = kwargs.pop("index", DEFAULT_INDEX)
        if self.index[-1] != "/":
            self.index += "/"
        super().__init__(*args, **kwargs)

    def _find(self):
        source_url = self.get_source_url()
        return CrateDiscoveredSource([source_url], timeout=self.timeout)

    def _get_release_history(self):
        url = f"{CRATES_API}{self.name}/versions"
        headers = {'User-Agent': 'soufi'}
        try:
            data = self.get_url(url, headers=headers).json()
        except exceptions.DownloadError:
            raise exceptions.SourceNotFound

        history = []
        for item in data.get('versions', []):
            if item.get('yanked'):
                continue
            version = item.get('num')
            if not version:
                continue
            history.append(
                {
                    'version': version,
                    'published_at': item.get('created_at'),
                }
            )
        if history == []:
            raise exceptions.SourceNotFound

        history.sort(
            key=lambda h: (
                h['published_at'] is None,
                h['published_at'] or "",
            )
        )
        return history

    def get_source_url(self):
        """Examine the index to find the source URL for the package.

        This is simply a matter of using the index's API to do a package query,
        and returning a computed URL for any exact match.

        Note: the Create index server's API does not do exact matches itself,
        so we need to iterate the results.
        """
        dl = self.get_index_dl()
        # This is a bit of a short cut and is assuming a particular file format
        # for each crate.  Nominally this is usually the case but we might
        # need to revist.
        url = f"{dl}/{self.name}/{self.name}-{self.version}.crate"
        if self.test_url(url):
            return url
        raise exceptions.SourceNotFound

    def get_index_dl(self):
        """Return the 'dl' value from the config.json at the index root."""
        # 'dl' is the URL prefix from which all downloads are made.
        config = self.get_url(f"{self.index}config.json").json()
        try:
            return config["dl"]
        except KeyError:
            raise exceptions.DownloadError(
                "Index is corrupt: No 'dl' key in index config.json"
            )


class CrateDiscoveredSource(finder.DiscoveredSource):
    """A discovered Rust Crate package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = ""  # .crate is part of the URL

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but Crates are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2024 Cisco Systems, Inc. and its affiliates
# All rights reserved.


import defusedxml.lxml
import requests

from soufi import exceptions, finder

DEFAULT_INDEX = "https://pecl.php.net/"


class PHPPECL(finder.SourceFinder):
    """Find PHP PECL packages.

    Looks up the package in the PECL index and returns the URL for the package.
    """

    distro = finder.SourceType.phppecl.value

    def _find(self):
        source_url = self.get_source_url()
        return PHPPECLDiscoveredSource([source_url], timeout=self.timeout)

    def get_source_url(self):
        """Examine the index to find the source URL for the package.

        This is simply a matter of using the index's REST API to do a package
        query, and returning a URL contained in the returned XML data.
        """
        url = f"{DEFAULT_INDEX}rest/r/{self.name}/{self.version}.xml"
        # The returned XML document contains a <g> element with the URL.
        try:
            with requests.get(url, stream=True, timeout=30) as r:
                r.raw.decode_content = True
                xml = defusedxml.lxml.parse(r.raw)
            return xml.find('.//{*}g').text
        except Exception:
            raise exceptions.SourceNotFound


class PHPPECLDiscoveredSource(finder.DiscoveredSource):
    """A discovered PHP PECL package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = ".tgz"

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but PECL archives are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2024 Cisco Systems, Inc. and its affiliates
# All rights reserved.


import requests

from soufi import exceptions, finder

DEFAULT_INDEX = "https://repo.packagist.org/"


class PHPComposer(finder.SourceFinder):
    """Find PHP Composer packages at the Packagist repo.

    Looks up the package in the Packagist repo and returns the URL for the
    package.
    """

    distro = finder.SourceType.phpcomposer.value

    def _find(self):
        source_url, extension = self.get_source_url()
        return PHPComposerDiscoveredSource(
            [source_url],
            timeout=self.timeout,
            archive_extension=f".{extension}",
        )

    def get_source_url(self):
        """Examine the repo to find the source URL for the package.

        Packagist has two APIs (see https://packagist.org/apidoc). The main one
        doesn't seem to allow queries for specific packages (nor at specific
        versions), so we can use repo.packagist.org optimistically by querying
        a known URL pattern at repo.packagist.org/p2/{package}.json. If the
        package exists, the metadata that is returned contains a list of all
        versions of the package that will need to be iterated to get the
        download URL.

        Note that the name of Composer packages usually takes the form of
        vendor/package, so the name of the package is the concatenation of
        these two parts.

        :return: A tuple of (URL, type) where the type is the archive_extension
            to use.
        """
        url = f"{DEFAULT_INDEX}p2/{self.name}.json"
        resp = requests.get(url, timeout=self.timeout)
        if resp.status_code != requests.codes.ok:
            raise exceptions.SourceNotFound

        try:
            versions = resp.json()["packages"][self.name]
            for version in versions:
                if self.version in (
                    version["version"],
                    version['version_normalized'],
                ):
                    return version["dist"]["url"], version["dist"]["type"]
        except Exception:
            raise exceptions.DownloadError(
                "Malformed JSON response from {url}"
            )

        raise exceptions.SourceNotFound


class PHPComposerDiscoveredSource(finder.DiscoveredSource):
    """A discovered PHP Composer package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive

    def __init__(self, *args, archive_extension, **kwargs):
        super().__init__(*args, **kwargs)
        self.archive_extension = archive_extension

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but PECL archives are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return f"{self.urls[0]}: {self.archive_extension}"



# Copyright (c) 2026 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from pathlib import Path

import fixtures
from testtools.matchers import Contains, Equals

from soufi import history_cli
from soufi.testing import base


class TestHistoryCLI(base.TestCase):
    def test_get_release_history_uses_disk_cache(self):
        history = [
            {
                "version": "1.0.0",
                "published_at": "2024-01-01T00:00:00Z",
            }
        ]
        source_finder = self.patch(history_cli.finder, "factory")
        source_finder.return_value.get_release_history.return_value = history
        temp_dir = self.useFixture(fixtures.TempDir()).path
        cache_path = str(Path(temp_dir) / "history-cache.dbm")

        result = history_cli.get_release_history(
            distro="python",
            name="example",
            cache_path=cache_path,
            cache_ttl=111,
            timeout=9,
            pyindex="https://pypi.org/pypi/",
            goproxy="https://proxy.golang.org/",
        )

        self.assertThat(result, Equals(history))
        _, kwargs = source_finder.call_args
        self.assertThat(kwargs["cache_backend"], Equals("dogpile.cache.dbm"))
        self.assertThat(kwargs["cache_ttl"], Equals(111))
        self.assertThat(kwargs["timeout"], Equals(9))
        self.assertThat(kwargs["cache_args"]["filename"], Contains(".dbm"))

    def test_get_release_history_rejects_unknown_type(self):
        temp_dir = self.useFixture(fixtures.TempDir()).path
        cache_path = str(Path(temp_dir) / "x.dbm")
        self.assertRaises(
            ValueError,
            history_cli.get_release_history,
            distro="nope",
            name="example",
            cache_path=cache_path,
            cache_ttl=1,
            timeout=1,
            pyindex="x",
            goproxy="x",
        )

    def test_get_release_history_passes_go_proxy_for_go(self):
        source_finder = self.patch(history_cli.finder, "factory")
        source_finder.return_value.get_release_history.return_value = []
        temp_dir = self.useFixture(fixtures.TempDir()).path
        cache_path = str(Path(temp_dir) / "history-cache.dbm")

        history_cli.get_release_history(
            distro="go",
            name="example/module",
            cache_path=cache_path,
            cache_ttl=111,
            timeout=9,
            pyindex="https://pypi.org/pypi/",
            goproxy="https://proxy.example.invalid/",
        )

        _, kwargs = source_finder.call_args
        self.assertThat(
            kwargs["goproxy"], Equals("https://proxy.example.invalid/")
        )



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

"""A simple testing CLI to find and download source."""

import functools
import json
import os
import shutil
import sys
import warnings

from soufi import exceptions, finder, history_cli

try:
    import click
    import pylru
except ImportError:
    sys.exit("CLI support not installed; please install soufi[cli]")

warnings.formatwarning = lambda msg, *x, **y: f"WARNING: {msg}\n"

# Configure a small-ish in-memory LRU cache to speed up operations
LRU_CACHE = pylru.lrucache(size=512)


class Finder:
    @classmethod
    def find(cls, finder):
        source = finder.find()
        click.echo(source)
        return source

    @classmethod
    def ubuntu(cls, name, version, timeout=None):
        click.echo("Logging in to Launchpad")
        ubuntu_finder = finder.factory(
            "ubuntu",
            name,
            version,
            finder.SourceType.os,
            cache_backend="dogpile.cache.memory",
            cache_args=dict(cache_dict=LRU_CACHE),
            timeout=timeout,
        )
        click.echo("Finding source in Launchpad")
        return cls.find(ubuntu_finder)

    @classmethod
    def debian(cls, name, version, timeout=None):
        debian_finder = finder.factory(
            "debian", name, version, finder.SourceType.os
        )
        return cls.find(debian_finder)

    @classmethod
    def npm(cls, name, version, timeout=None):
        npm_finder = finder.factory(
            "npm", name, version, finder.SourceType.npm, timeout=timeout
        )
        return cls.find(npm_finder)

    @classmethod
    def python(cls, name, version, pyindex=None, timeout=None):
        python_finder = finder.factory(
            "python",
            name=name,
            version=version,
            s_type=finder.SourceType.python,
            pyindex=pyindex,
            timeout=timeout,
        )
        return cls.find(python_finder)

    @classmethod
    def centos(
        cls,
        name,
        version,
        repos=None,
        source_repos=None,
        binary_repos=None,
        timeout=None,
    ):
        optimal = "optimal" in repos
        centos_finder = finder.factory(
            "centos",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            repos=repos,
            optimal_repos=optimal,
            source_repos=source_repos,
            binary_repos=binary_repos,
            cache_backend="dogpile.cache.memory",
            cache_args=dict(cache_dict=LRU_CACHE),
            timeout=timeout,
        )
        return cls.find(centos_finder)

    @classmethod
    def almalinux(
        cls,
        name,
        version,
        repos=None,
        source_repos=None,
        binary_repos=None,
        timeout=None,
    ):
        alma_finder = finder.factory(
            "almalinux",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            source_repos=source_repos,
            binary_repos=binary_repos,
            cache_backend="dogpile.cache.memory",
            cache_args=dict(cache_dict=LRU_CACHE),
            timeout=timeout,
        )
        return cls.find(alma_finder)

    @classmethod
    def alpine(cls, name, version, aports_dir, timeout=None):
        alpine_finder = finder.factory(
            "alpine",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            aports_dir=aports_dir,
            cache_backend="dogpile.cache.memory",
            cache_args=dict(cache_dict=LRU_CACHE),
            timeout=timeout,
        )
        return cls.find(alpine_finder)

    @classmethod
    def go(cls, name, version, goproxy, timeout=None):
        go_finder = finder.factory(
            "go",
            name=name,
            version=version,
            s_type=finder.SourceType.go,
            goproxy=goproxy,
            timeout=timeout,
        )
        return cls.find(go_finder)

    @classmethod
    def java(cls, name, version, timeout=None):
        java_finder = finder.factory(
            "java",
            name=name,
            version=version,
            s_type=finder.SourceType.java,
            timeout=timeout,
        )
        return cls.find(java_finder)

    @classmethod
    def gem(cls, name, version, timeout=None):
        gem_finder = finder.factory(
            "gem",
            name=name,
            version=version,
            s_type=finder.SourceType.gem,
            timeout=timeout,
        )
        return cls.find(gem_finder)

    @classmethod
    def photon(
        cls, name, version, source_repos=None, binary_repos=None, timeout=None
    ):
        photon_finder = finder.factory(
            "photon",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            source_repos=source_repos,
            binary_repos=binary_repos,
            cache_backend="dogpile.cache.memory",
            cache_args=dict(cache_dict=LRU_CACHE),
            timeout=timeout,
        )
        return cls.find(photon_finder)

    @classmethod
    def rhel(
        cls, name, version, source_repos=None, binary_repos=None, timeout=None
    ):
        rhel_finder = finder.factory(
            "rhel",
            name=name,
            version=version,
            s_type=finder.SourceType.os,
            source_repos=source_repos,
            binary_repos=binary_repos,
            cache_backend="dogpile.cache.memory",
            cache_args=dict(cache_dict=LRU_CACHE),
            timeout=timeout,
        )
        return cls.find(rhel_finder)

    @classmethod
    def crate(cls, name, version, timeout=None):
        crate_finder = finder.factory(
            "crate",
            name=name,
            version=version,
            s_type=finder.SourceType.crate,
            timeout=timeout,
        )
        return cls.find(crate_finder)

    @classmethod
    def phppecl(cls, name, version, timeout=None):
        pecl_finder = finder.factory(
            "phppecl",
            name=name,
            version=version,
            s_type=finder.SourceType.phppecl,
            timeout=timeout,
        )
        return cls.find(pecl_finder)

    @classmethod
    def phpcomposer(cls, name, version, timeout=None):
        composer_finder = finder.factory(
            "phpcomposer",
            name=name,
            version=version,
            s_type=finder.SourceType.phpcomposer,
            timeout=timeout,
        )
        return cls.find(composer_finder)


class DefaultFindGroup(click.Group):
    """A command group that defaults to the legacy implicit `find` mode."""

    def resolve_command(self, ctx, args):
        if (
            args
            and args[0] not in self.commands
            and not args[0].startswith("-")
        ):
            ctx.meta["implicit_find_mode"] = True
            args = ["find", *args]
        return super().resolve_command(ctx, args)


def make_archive_from_discovery_source(disc_src, fname):
    try:
        with disc_src.make_archive() as in_fd, open(fname, "wb") as out_fd:
            # copyfileobj copies in chunks, so as not to exhaust memory.
            shutil.copyfileobj(in_fd, out_fd)
    except exceptions.DownloadError as e:
        click.echo(str(e))
        click.get_current_context().exit(255)


@click.group(cls=DefaultFindGroup)
def main():
    """Find source files and query upstream release history."""


@main.command("find")
@click.argument("distro")
@click.argument("name")
@click.argument("version")
@click.option(
    "--pyindex",
    default="https://pypi.org/pypi/",
    help="Python package index if getting Python",
    show_default=True,
)
@click.option(
    "--aports",
    default=None,
    help="Path to a checked-out aports directory if getting Alpine",
    show_default=True,
)
@click.option(
    "--repo",
    default=(),
    multiple=True,
    help="For CentOS/Almalinux, name of repo to use instead of defaults. "
    "Use 'optimal' to use an extended optimal set. May be repeated.",
)
@click.option(
    "--source-repo",
    default=(),
    multiple=True,
    help="For Yum-based distros, URL of a source repo mirror to use "
    "for lookups instead of the distro defaults. May be repeated. "
    "On CentOS, this causes --repo to be ignored.",
)
@click.option(
    "--binary-repo",
    default=(),
    multiple=True,
    help="For Yum-based distros, URL of a binary repo mirror to use "
    "for lookups instead of the distro defaults. May be repeated. "
    "On CentOS, this causes --repo to be ignored.",
)
@click.option(
    "--goproxy",
    default="https://proxy.golang.org/",
    help="GOPROXY to use when downloading Golang module source",
    show_default=True,
)
@click.option(
    "--output",
    "-o",
    help="Download the source archive and write to this file name",
    default=None,
)
@click.option(
    "-O",
    "auto_output",
    is_flag=True,
    default=False,
    help="Download the source archive and write to a default file name."
    "The names vary according to the source type and the archive type, "
    "but will generally follow the format: "
    "{name}-{version}.{distro/type}.tar.[gz|xz] "
    "(This option takes precedence over -o/--output)",
)
@click.option(
    "--timeout",
    type=click.IntRange(1),
    default=30,
    help="Timeout when making requests.  Must be greater than zero.",
    show_default=True,
)
def find(
    distro,
    name,
    version,
    pyindex,
    aports,
    repo,
    goproxy,
    output,
    auto_output,
    source_repo,
    binary_repo,
    timeout,
):
    """Find and optionally download source files.

    Given a binary name and version, will find and print the URL(s) to the
    source file(s).

    If the --output option is present, the URLs are all downloaded and
    combined into a LZMA-compressed tar file and written to the file
    name specifed.  If the original source is already an archive then that
    archive is used instead.

    The sources currently supported are 'debian', 'ubuntu', 'rhel', 'centos',
    'alpine', 'photon', 'java', 'go', 'python', 'create', 'phppecl',
    'phpcomposer', and 'npm', one of which must be specified as the DISTRO
    argument.
    """
    if click.get_current_context().meta.get("implicit_find_mode", False):
        click.echo(
            "WARNING: `soufi ...` without `find` is deprecated and will be "
            "removed in a future release; use `soufi find ...` instead.",
            err=True,
        )
    try:
        func = functools.partial(
            getattr(Finder, distro), name, version, timeout=timeout
        )
    except AttributeError:
        click.echo(f"{distro} not available")
        click.get_current_context().exit(255)
    if distro == "alpine" and aports is None:
        click.echo("Must provide --aports for Alpine")
        click.get_current_context().exit(255)
    try:
        if distro == "python":
            disc_source = func(pyindex=pyindex)
        elif distro == "alpine":
            disc_source = func(aports_dir=aports)
        elif distro == "centos":
            disc_source = func(
                repos=repo, source_repos=source_repo, binary_repos=binary_repo
            )
        elif distro in ("photon", "rhel"):
            disc_source = func(
                source_repos=source_repo, binary_repos=binary_repo
            )
        elif distro == "go":
            disc_source = func(goproxy=goproxy)
        else:
            disc_source = func()
    except exceptions.SourceNotFound:
        click.echo("source not found")
        click.get_current_context().exit(255)

    if auto_output is not False or output is not None:
        fname = output
        if auto_output:
            name = name.replace(os.sep, ".")
            fname = f"{name}-{version}.{distro}{disc_source.archive_extension}"
        make_archive_from_discovery_source(disc_source, fname)


@main.command("history")
@click.argument("distro", type=click.Choice(sorted(history_cli.HISTORY_TYPES)))
@click.argument("name")
@click.option(
    "--cache-path",
    default=history_cli.DEFAULT_CACHE_PATH,
    show_default=True,
    help="On-disk cache path for release history",
)
@click.option(
    "--cache-ttl",
    default=history_cli.DEFAULT_CACHE_TTL,
    type=click.IntRange(0),
    show_default=True,
    help="Cache TTL in seconds",
)
@click.option(
    "--timeout",
    default=30,
    type=click.IntRange(1),
    show_default=True,
    help="HTTP request timeout in seconds",
)
@click.option(
    "--pyindex",
    default="https://pypi.org/pypi/",
    show_default=True,
    help="Python package index URL",
)
@click.option(
    "--goproxy",
    default="https://proxy.golang.org/",
    show_default=True,
    help="Go proxy URL",
)
@click.option(
    "--json",
    "as_json",
    is_flag=True,
    default=False,
    help="Render output as JSON",
)
def history(
    distro,
    name,
    cache_path,
    cache_ttl,
    timeout,
    pyindex,
    goproxy,
    as_json,
):
    """Return release history for a package ecosystem/type + name."""
    try:
        data = history_cli.get_release_history(
            distro=distro,
            name=name,
            cache_path=cache_path,
            cache_ttl=cache_ttl,
            timeout=timeout,
            pyindex=pyindex,
            goproxy=goproxy,
        )
    except (
        ValueError,
        exceptions.SourceNotFound,
        exceptions.DownloadError,
    ) as e:
        click.echo(str(e), err=True)
        click.get_current_context().exit(255)

    if as_json:
        click.echo(json.dumps(data, indent=2, sort_keys=True))
        return
    for item in data:
        version = item.get("version", "")
        published_at = item.get("published_at") or "-"
        click.echo(f"{version}\t{published_at}")


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

import requests
from testtools.matchers._basic import SameMembers

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import golang
from soufi.testing import base


class TestGolangFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        kwargs = dict(
            name=name,
            version=version,
            s_type=SourceType.go,
        )
        return golang.GolangFinder(**kwargs)

    def test_raises_when_module_not_found(self):
        finder = self.make_finder()
        self.patch_head_with_response(requests.codes.not_found)
        self.assertRaises(exceptions.SourceNotFound, finder.find)

    def test_finds_module(self):
        finder = self.make_finder()
        head = self.patch_head_with_response(requests.codes.ok)
        source = finder.find()
        expected = [
            f"{golang.PUBLIC_PROXY}{finder.name.lower()}"
            f"/@v/{finder.version}.zip"
        ]
        self.assertThat(source.urls, SameMembers(expected))
        head.assert_called_once_with(
            expected[0], timeout=finder.timeout, allow_redirects=True
        )

    def test_retries_with_get_if_head_fails(self):
        finder = self.make_finder()
        head = self.patch_head_with_response(requests.codes.not_allowed)
        get = self.patch_get_with_response(requests.codes.ok)
        source = finder.find()
        expected = [
            f"{golang.PUBLIC_PROXY}{finder.name.lower()}"
            f"/@v/{finder.version}.zip"
        ]
        self.assertThat(source.urls, SameMembers(expected))
        head.assert_called_once_with(
            expected[0], timeout=finder.timeout, allow_redirects=True
        )
        get.assert_called_once_with(
            expected[0], stream=True, timeout=finder.timeout
        )

    def test_get_release_history(self):
        finder = self.make_finder()
        get = self.patch_get_with_response(requests.codes.ok)
        get.return_value.text = "v1.0.0\nv1.1.0\n"

        history = finder.get_release_history()
        self.assertEqual(
            ['v1.0.0', 'v1.1.0'], [item['version'] for item in history]
        )

    def test_get_release_history_reverses_when_latest_first(self):
        finder = self.make_finder()
        get = self.patch(requests, 'get')

        def fake_get(url, *args, **kwargs):
            response = mock.MagicMock()
            response.status_code = requests.codes.ok
            if url.endswith('/@v/list'):
                response.text = "v1.1.0\nv1.0.0\n"
            else:
                response.json.return_value = {'Version': 'v1.1.0'}
            return response

        get.side_effect = fake_get
        history = finder.get_release_history()
        self.assertEqual(
            ['v1.0.0', 'v1.1.0'], [item['version'] for item in history]
        )

    def test_get_release_history_raises_when_request_fails(self):
        finder = self.make_finder()
        self.patch_get_with_response(requests.codes.bad_request)
        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )

    def test_get_release_history_raises_when_empty(self):
        finder = self.make_finder()
        get = self.patch_get_with_response(requests.codes.ok)
        get.return_value.text = "\n"
        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )

    def test_get_release_history_handles_latest_download_error(self):
        finder = self.make_finder()
        get = self.patch(requests, 'get')

        def fake_get(url, *args, **kwargs):
            response = mock.MagicMock()
            response.status_code = requests.codes.ok
            if url.endswith('/@v/list'):
                response.text = "v1.0.0\nv1.1.0\n"
            else:
                response.raise_for_status.side_effect = requests.HTTPError
            return response

        get.side_effect = fake_get
        history = finder.get_release_history()
        self.assertEqual(
            ['v1.0.0', 'v1.1.0'], [item['version'] for item in history]
        )

    def test_get_latest_version_returns_none_on_download_error(self):
        finder = self.make_finder()
        get_url = self.patch(finder, 'get_url')
        get_url.side_effect = exceptions.DownloadError("boom")
        self.assertIs(None, finder._get_latest_version())


class TestGolangDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return golang.GolangDiscoveredSource([url])

    def test_make_archive(self):
        gds = self.make_discovered_source()
        self.assertEqual(gds.make_archive, gds.remote_url_is_archive)

    def test_repr(self):
        url = self.factory.make_url()
        gds = self.make_discovered_source(url=url)
        self.assertEqual(url, repr(gds))



# Copyright (c) 2024 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import json
from pathlib import Path
from unittest import mock

import requests

from soufi import exceptions, testing
from soufi.finder import SourceType
from soufi.finders import php_composer
from soufi.testing import base


class TestPHPPECLFinder(base.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.testing = Path(testing.__path__[0]) / 'data' / 'phpcomposer'

    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string("name")
        if version is None:
            version = self.factory.make_string("version")
        kwargs = dict(
            name=name, version=version, s_type=SourceType.phpcomposer
        )
        return php_composer.PHPComposer(**kwargs)

    def test_get_source_url(self):
        # The test data uses this package.
        finder = self.make_finder("monolog/monolog", "2.3.5")
        expected_url = self.factory.make_url()
        expected_type = self.factory.random_choice(["zip", "tgz", "tar.gz"])
        with open(self.testing / "monolog.json", "rb") as fp:
            data = json.load(fp)
        # Patch the real URL with our random one to make sure it's correctly
        # picked up later.
        for version in data["packages"]["monolog/monolog"]:
            if version["version"] == "2.3.5":
                version["dist"]["url"] = expected_url
                version["dist"]["type"] = expected_type
                break
        get = self.patch_get_with_response(requests.codes.ok, json=data)

        found_url, found_type = finder.get_source_url()
        self.assertEqual(expected_url, found_url)
        self.assertEqual(expected_type, found_type)
        call = mock.call(
            f"{php_composer.DEFAULT_INDEX}p2/{finder.name}.json",
            timeout=30,
        )
        self.assertIn(call, get.call_args_list)

    def test_get_source_url_source_not_found(self):
        finder = self.make_finder()
        self.patch_get_with_response(requests.codes.not_found)
        self.assertRaises(exceptions.SourceNotFound, finder.get_source_url)

    def test_get_source_version_not_found(self):
        finder = self.make_finder("monolog/monolog")
        with open(self.testing / "monolog.json", "rb") as fp:
            data = json.load(fp)
        self.patch_get_with_response(requests.codes.ok, json=data)
        self.assertRaises(exceptions.SourceNotFound, finder.get_source_url)

    def test_get_source_url_malformed_json(self):
        finder = self.make_finder()
        self.patch_get_with_response(requests.codes.ok, data="not json")
        self.assertRaises(exceptions.DownloadError, finder.get_source_url)

    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        atype = self.factory.random_choice(["zip", "tgz", "tar.gz"])
        self.patch(finder, "get_source_url").return_value = url, atype

        disc_source = finder.find()
        self.assertIsInstance(
            disc_source, php_composer.PHPComposerDiscoveredSource
        )
        self.assertEqual([url], disc_source.urls)
        self.assertEqual(f".{atype}", disc_source.archive_extension)


class TestPHPPECLDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None, atype=None):
        if url is None:
            url = self.factory.make_url()
        if atype is None:
            atype = self.factory.random_choice([".zip", ".tgz", ".tar.gz"])
        return php_composer.PHPComposerDiscoveredSource(
            [url], archive_extension=atype
        )

    def test_repr(self):
        url = self.factory.make_url()
        atype = self.factory.random_choice([".zip", ".tgz", ".tar.gz"])
        ds = self.make_discovered_source(url, atype)
        self.assertEqual(f"{url}: {atype}", repr(ds))

    def test_make_archive(self):
        ds = self.make_discovered_source()
        self.assertEqual(ds.make_archive, ds.remote_url_is_archive)



# Copyright (c) 2021-2023 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import requests
import testtools

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import npm
from soufi.testing import base


class TestNPMFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return npm.NPMFinder(name, version, SourceType.npm)

    def test_get_source_url(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        data = dict(dist=dict(tarball=url))

        get = self.patch_get_with_response(requests.codes.ok, json=data)
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        get.assert_called_once_with(
            f"https://registry.npmjs.org/{finder.name}/{finder.version}",
            timeout=finder.timeout,
        )

    def test_get_source_info_raises_when_response_fails(self):
        self.patch_get_with_response(requests.codes.not_found)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_find(self):
        url = self.factory.make_url()
        finder = self.make_finder()
        self.patch(finder, 'get_source_url').return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, npm.NPMDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test_get_release_history(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'versions': {'1.0.0': {}, '2.0.0': {}},
            'time': {
                'created': '2020-01-01T00:00:00.000Z',
                'modified': '2025-01-01T00:00:00.000Z',
                '1.0.0': '2024-01-01T00:00:00.000Z',
                '2.0.0': '2024-02-01T00:00:00.000Z',
            },
        }
        self.patch_get_with_response(requests.codes.ok, json=data)

        history = finder.get_release_history()
        self.assertEqual(
            ['1.0.0', '2.0.0'], [item['version'] for item in history]
        )

    def test_get_release_history_raises_when_request_fails(self):
        finder = self.make_finder(name=self.factory.make_string())
        self.patch_get_with_response(requests.codes.bad_request)
        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )

    def test_get_release_history_falls_back_to_versions_when_time_empty(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'versions': {'1.0.0': {}, '2.0.0': {}},
            'time': {},
        }
        self.patch_get_with_response(requests.codes.ok, json=data)

        history = finder.get_release_history()
        self.assertEqual(
            ['1.0.0', '2.0.0'], sorted(item['version'] for item in history)
        )

    def test_get_release_history_raises_when_no_versions(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'versions': {},
            'time': {'2.0.0': '2024-02-01T00:00:00.000Z'},
        }
        self.patch_get_with_response(requests.codes.ok, json=data)
        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )


class TestNPMDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return npm.NPMDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        nds = self.make_discovered_source(url)
        self.assertEqual(url, repr(nds))

    def test_make_archive(self):
        nds = self.make_discovered_source()
        self.assertEqual(nds.make_archive, nds.remote_url_is_archive)



# Copyright (c) 2026 Cisco Systems, Inc. and its affiliates
# All rights reserved.

"""Simple CLI for package release history lookups."""

from __future__ import annotations

import os
from typing import Any

from soufi import finder

DEFAULT_CACHE_PATH = "~/.cache/soufi-history.dbm"
DEFAULT_CACHE_TTL = 60 * 60 * 24

HISTORY_TYPES = {
    "python": finder.SourceType.python,
    "npm": finder.SourceType.npm,
    "gem": finder.SourceType.gem,
    "crate": finder.SourceType.crate,
    "go": finder.SourceType.go,
    "java": finder.SourceType.java,
}


def get_release_history(
    distro: str,
    name: str,
    cache_path: str,
    cache_ttl: int,
    timeout: int,
    pyindex: str,
    goproxy: str,
) -> list[dict[str, Any]]:
    """Return release history for the requested package."""
    if distro not in HISTORY_TYPES:
        raise ValueError(f"Unsupported distro/type: {distro}")

    cache_file = os.path.expanduser(cache_path)
    os.makedirs(os.path.dirname(cache_file) or ".", exist_ok=True)
    kwargs: dict[str, Any] = {
        "name": name,
        "s_type": HISTORY_TYPES[distro],
        "cache_backend": "dogpile.cache.dbm",
        "cache_args": {"filename": cache_file},
        "cache_ttl": cache_ttl,
        "timeout": timeout,
    }
    if distro == "python":
        kwargs["pyindex"] = pyindex
    if distro == "go":
        kwargs["goproxy"] = goproxy
    source_finder = finder.factory(distro, **kwargs)
    return source_finder.get_release_history()



# Copyright (c) 2026 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

from click.testing import CliRunner
from testtools.matchers import Contains, Not

from soufi import cli
from soufi.testing import base


class TestCLI(base.TestCase):
    def setUp(self):
        super().setUp()
        self.runner = CliRunner()

    def test_main_uses_legacy_implicit_find_mode(self):
        python = self.patch(cli.Finder, "python")
        python.return_value = object()

        result = self.runner.invoke(cli.main, ["python", "foo", "1.2.3"])

        self.assertEqual(0, result.exit_code, result.output)
        python.assert_called_once_with(
            "foo", "1.2.3", timeout=30, pyindex=mock.ANY
        )
        self.assertThat(result.output, Contains("deprecated"))

    def test_find_command_still_works_without_deprecation_warning(self):
        python = self.patch(cli.Finder, "python")
        python.return_value = object()

        result = self.runner.invoke(
            cli.main, ["find", "python", "foo", "1.2.3"]
        )

        self.assertEqual(0, result.exit_code, result.output)
        python.assert_called_once_with(
            "foo", "1.2.3", timeout=30, pyindex=mock.ANY
        )
        self.assertThat(result.output, Not(Contains("deprecated")))

    def test_help_shows_find_subcommand(self):
        result = self.runner.invoke(cli.main, ["--help"])

        self.assertEqual(0, result.exit_code, result.output)
        self.assertThat(result.output, Contains("\n  find"))
        self.assertThat(result.output, Contains("history"))



# Copyright (c) 2024 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

import requests

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import crate
from soufi.testing import base


class TestCrateFinder(base.TestCase):
    def make_finder(self, name=None, version=None, index=None):
        if name is None:
            name = self.factory.make_string("name")
        if version is None:
            version = self.factory.make_string("version")
        kwargs = dict(name=name, version=version, s_type=SourceType.crate)
        if index is not None:
            kwargs["index"] = index
        return crate.CrateFinder(**kwargs)

    def test_get_source_url(self):
        finder = self.make_finder()
        get_index_dl = self.patch(finder, "get_index_dl")
        index_dl = self.factory.make_url()
        get_index_dl.return_value = index_dl

        head = self.patch_head_with_response(requests.codes.ok)
        found_url = finder.get_source_url()
        expected_url = (
            f"{index_dl}/{finder.name}/{finder.name}-{finder.version}.crate"
        )
        self.assertEqual(expected_url, found_url)
        call = mock.call(
            expected_url,
            timeout=30,
            allow_redirects=True,
        )
        self.assertIn(call, head.call_args_list)

    def test_get_source_url_source_not_found(self):
        finder = self.make_finder()
        self.patch(
            finder, "get_index_dl"
        ).return_value = self.factory.make_url()
        # Crates index returns a 403 when the crate is not found. ¯\_(ツ)_/¯
        self.patch_head_with_response(requests.codes.forbidden)
        self.assertRaises(exceptions.SourceNotFound, finder.get_source_url)

    def test_get_index_dl(self):
        index = self.factory.make_url()
        finder = self.make_finder(index=index)
        get_url = self.patch(finder, "get_url")
        index = self.factory.make_url()
        get_url.return_value.json.return_value = {"dl": index}

        found_index = finder.get_index_dl()
        self.assertEqual(index, found_index)
        call = mock.call(f"{finder.index}config.json")
        self.assertIn(call, get_url.call_args_list)

    def test_get_index_dl_no_dl_key(self):
        finder = self.make_finder()
        get_url = self.patch(finder, "get_url")
        get_url.return_value.json.return_value = {}

        self.assertRaises(exceptions.DownloadError, finder.get_index_dl)

    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, "get_source_url").return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, crate.CrateDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test_get_release_history(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'versions': [
                {'num': '2.0.0', 'created_at': '2024-02-01T00:00:00.000Z'},
                {'num': '1.0.0', 'created_at': '2024-01-01T00:00:00.000Z'},
            ]
        }
        get = self.patch_get_with_response(requests.codes.ok, json=data)

        history = finder.get_release_history()
        self.assertEqual(
            ['1.0.0', '2.0.0'], [item['version'] for item in history]
        )
        get.assert_called_with(
            f"{crate.CRATES_API}{finder.name}/versions",
            headers={'User-Agent': 'soufi'},
            timeout=finder.timeout,
        )

    def test_get_release_history_raises_when_request_fails(self):
        finder = self.make_finder(name=self.factory.make_string())
        self.patch_get_with_response(requests.codes.bad_request)

        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )

    def test_get_release_history_skips_yanked_and_empty_versions(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'versions': [
                {'num': '3.0.0', 'yanked': True},
                {'created_at': '2024-03-01T00:00:00.000Z'},
                {'num': '2.0.0', 'created_at': '2024-02-01T00:00:00.000Z'},
            ]
        }
        self.patch_get_with_response(requests.codes.ok, json=data)

        history = finder.get_release_history()
        self.assertEqual(['2.0.0'], [item['version'] for item in history])

    def test_get_release_history_raises_when_all_versions_filtered(self):
        finder = self.make_finder(name=self.factory.make_string())
        data = {
            'versions': [
                {'num': '3.0.0', 'yanked': True},
                {'created_at': '2024-03-01T00:00:00.000Z'},
            ]
        }
        self.patch_get_with_response(requests.codes.ok, json=data)

        self.assertRaises(
            exceptions.SourceNotFound,
            finder.get_release_history,
        )


class TestCrateDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return crate.CrateDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        pds = self.make_discovered_source(url)
        self.assertEqual(url, repr(pds))

    def test_make_archive(self):
        cds = self.make_discovered_source()
        self.assertEqual(cds.make_archive, cds.remote_url_is_archive)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.


from soufi import exceptions, finder

GEM_DOWNLOADS = 'https://rubygems.org/downloads/'
GEM_VERSIONS_URL = 'https://rubygems.org/api/v1/versions/'


class GemFinder(finder.SourceFinder):
    """Find Gem files.

    Finds files at https://rubygems.org/downloads
    """

    distro = finder.SourceType.gem.value

    def _find(self):
        source_url = self.get_source_url()
        return GemDiscoveredSource([source_url], timeout=self.timeout)

    def _get_release_history(self):
        url = f'{GEM_VERSIONS_URL}{self.name}.json'
        try:
            data = self.get_url(url).json()
        except exceptions.DownloadError:
            raise exceptions.SourceNotFound

        history = []
        for item in data:
            version = item.get('number')
            if not version:
                continue
            history.append(
                {
                    'version': version,
                    'published_at': item.get('created_at'),
                }
            )
        if history == []:
            raise exceptions.SourceNotFound

        history.sort(
            key=lambda h: (
                h['published_at'] is None,
                h['published_at'] or "",
            )
        )
        return history

    def get_source_url(self):
        url = f"{GEM_DOWNLOADS}{self.name}-{self.version}.gem"
        if not self.test_url(url):
            raise exceptions.SourceNotFound
        return url


class GemDiscoveredSource(finder.DiscoveredSource):
    """A discovered Gem package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    # NOTE(nic): `distro` and `archive_extension` are the same,
    # so auto-output mode will name these things `.gem.gem`.  The
    # recommended workaround is to not use auto-output mode :-)
    archive_extension = '.gem'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but Gems are already tarballs so
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.


from soufi import exceptions, finder

MAVEN_SEARCH_URL = 'https://search.maven.org/solrsearch/select'
MAVEN_REPO_URL = 'https://search.maven.org/remotecontent'


class JavaFinder(finder.SourceFinder):
    """Find Java source JARs.

    Uses the SOLR search at https://search.maven.org/

    :param group_id: Optional Maven groupId to narrow the search. Without
        this, the search may return versions from different artifacts that
        share the same artifactId.
    """

    distro = finder.SourceType.java.value

    def __init__(self, *args, group_id=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.group_id = group_id

    def _find(self):
        source_url = self.get_source_url()
        return JavaDiscoveredSource([source_url], timeout=self.timeout)

    def _get_release_history(self):
        # Build query - if group_id is provided, filter by it to avoid
        # matching different artifacts with the same name.
        if self.group_id:
            query = f'g:{self.group_id} AND a:{self.name} l:sources'
        else:
            query = f'a:{self.name} l:sources'
        params = dict(
            q=query,
            rows=200,
            core='gav',
            wt='json',
            sort='timestamp asc',
        )
        try:
            data = self.get_url(MAVEN_SEARCH_URL, params=params).json()
        except exceptions.DownloadError:
            raise exceptions.SourceNotFound

        docs = data.get('response', {}).get('docs', [])
        history = []
        seen = set()
        for doc in docs:
            version = doc.get('v')
            if not version or version in seen:
                continue
            seen.add(version)
            timestamp = doc.get('timestamp')
            if timestamp is not None:
                # Search returns timestamp in milliseconds since epoch.
                timestamp = f"{int(timestamp)}"
            history.append(
                {
                    'version': version,
                    'published_at': timestamp,
                }
            )

        if history == []:
            raise exceptions.SourceNotFound

        history.sort(
            key=lambda h: (
                h['published_at'] is not None,
                h['published_at'] or "",
            )
        )
        return history

    def get_source_url(self):
        """Construct a URL from the JSON response for the search."""
        params = dict(q=f'a:{self.name} v:{self.version} l:sources', rows=1)
        # NOTE(nic): empirical observation of throwing "bad" queries at this
        #  thing lead me to believe this will only ever return 200 OK unless
        #  the query string is literally corrupted; if you pass in a
        #  syntactically valid search string for a resource that does not
        #  exist, the response content simply contains a completely
        #  unrelated record.  :-(
        try:
            data = self.get_url(MAVEN_SEARCH_URL, params=params).json()
        except exceptions.DownloadError:
            raise exceptions.SourceNotFound

        # Transform the `group` attribute and use it to put together a
        # resource URL from the other data we have.  If it resolves, return
        # it, otherwise the previous search was no good.
        [record] = data['response']['docs']
        group = record['g'].replace('.', '/')
        params = dict(
            filepath=f'{group}/{self.name}/{self.version}/'
            f'{self.name}-{self.version}-sources.jar'
        )
        found = self.test_url(
            MAVEN_REPO_URL,
            params=params,
        )
        if not found:
            raise exceptions.SourceNotFound
        return found.url


class JavaDiscoveredSource(finder.DiscoveredSource):
    """A discovered Java source package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = '.jar'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Required by the base class but JARs are already ZIP files so there is
        # nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]



# Copyright (c) 2024 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from pathlib import Path
from unittest import mock

import requests

from soufi import exceptions, testing
from soufi.finder import SourceType
from soufi.finders import php_pecl
from soufi.testing import base


class TestPHPPECLFinder(base.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.testing = Path(testing.__path__[0]) / 'data' / 'phppecl'

    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string("name")
        if version is None:
            version = self.factory.make_string("version")
        kwargs = dict(name=name, version=version, s_type=SourceType.phppecl)
        return php_pecl.PHPPECL(**kwargs)

    def test_get_source_url(self):
        # The test data uses this package at this version.
        finder = self.make_finder("ncurses", "1.0.2")
        get = self.patch_get_with_response(requests.codes.ok)
        fp = open(self.testing / "ncurses_1.0.2.xml", "rb")
        self.addCleanup(fp.close)
        # Patch the get context manager to return the file stream.
        get.return_value.__enter__.return_value.raw = fp

        found_url = finder.get_source_url()
        expected_url = (
            f"{php_pecl.DEFAULT_INDEX}get/{finder.name}-{finder.version}"
        )
        self.assertEqual(expected_url, found_url)
        call = mock.call(
            f"{php_pecl.DEFAULT_INDEX}rest/r/"
            f"{finder.name}/{finder.version}.xml",
            stream=True,
            timeout=30,
        )
        self.assertIn(call, get.call_args_list)

    def test_get_source_url_source_not_found(self):
        finder = self.make_finder()
        self.patch_get_with_response(requests.codes.not_found)
        self.assertRaises(exceptions.SourceNotFound, finder.get_source_url)

    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, "get_source_url").return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, php_pecl.PHPPECLDiscoveredSource)
        self.assertEqual([url], disc_source.urls)


class TestPHPPECLDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return php_pecl.PHPPECLDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        ds = self.make_discovered_source(url)
        self.assertEqual(url, repr(ds))

    def test_make_archive(self):
        ds = self.make_discovered_source()
        self.assertEqual(ds.make_archive, ds.remote_url_is_archive)
